//
// The provider for a tree of ECMA documents
//
// Author:
//   Miguel de Icaza (miguel@ximian.com)
//
// (C) 2002, 2003 Ximian, Inc.
//
// TODO:
//   Should cluster together constructors
//   Should cluster together methods with the same name (overloads)
//
// Easy:
//   Should render attributes on the signature.
//   Include examples as well.
//
using System;
using System.Reflection;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Text;
using System.Collections;
using ICSharpCode.SharpZipLib.Zip;

using H = HtmlHelper;

//
// Helper routines to extract information from an Ecma XML document
//
public class EcmaDoc {
	public static string GetFullClassName (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type").Attributes ["FullName"].InnerText;
	}
	
	public static string GetClassName (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type").Attributes ["Name"].InnerText;
	}

	public static string GetClassNamespace (XmlDocument doc)
	{
		string s = doc.SelectSingleNode ("/Type").Attributes ["FullName"].InnerText;

		return s.Substring (0, s.LastIndexOf ("."));
	}
	
	public static string GetTypeKind (XmlDocument doc)
	{
		XmlNode node = doc.SelectSingleNode ("/Type/Base/BaseTypeName");

		if (node == null){
			if (GetFullClassName (doc) == "System.Object")
				return "Class";
			return "Interface";
		}

		switch (node.InnerText){

		case "System.Delegate":
			return "Delegate";
		case "System.ValueType":
			return "Structure";
		case "System.Enum":
			return "Enumeration";
		default:
			return "Class";
		}
	}

	///
	/// <method>
	///  Returns the arguments of a node, including the parenthesis
	/// </method>
	///
	public static string GetArguments (XmlNode doc)
	{
		StringBuilder builder = new StringBuilder ();
		XmlNodeList parameters = doc.SelectNodes ("Parameters/Parameter");

		if (parameters.Count == 0)
			return null;

		else {
			builder.Append ("(");
			for (int i = 0; i < parameters.Count; i++) {
				builder.Append (
					ConvertCTSName (parameters [i].Attributes ["Type"].Value));

				if (i + 1 < parameters.Count)
					builder.Append (", ");
			}
		}

		builder.Append (")");
		return builder.ToString ();
	}
	
	public static string GetAssembly (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type/AssemblyInfo/AssemblyName").InnerText;
	}

	//
	// Utility function: converts a fully .NET qualified type name into a C#-looking one
	//
	static string ConvertCTSName (string type)
	{
		string retval = String.Empty;
		bool isArray = false;
		
		if (!type.StartsWith ("System."))
			return type;

		if (type.EndsWith ("[]")) {
			isArray = true;
			type = type.Substring (0, type.Length - 2);
		}
		
		switch (type) {
		case "System.Byte": retval = "byte"; break;
		case "System.SByte": retval = "sbyte"; break;
		case "System.Int16": retval = "short"; break;
		case "System.Int32": retval = "int"; break;
		case "System.Int64": retval = "long"; break;
			
		case "System.UInt16": retval = "ushort"; break;
		case "System.UInt32": retval = "uint"; break;
		case "System.UInt64": retval = "ulong"; break;
			
		case "System.Single":  retval = "float"; break;
		case "System.Double":  retval = "double"; break;
		case "System.Decimal": retval = "decimal"; break;
		case "System.Boolean": retval = "bool"; break;
		case "System.Char":    retval = "char"; break;
		case "System.String":  retval = "string"; break;
			
		default:
			if (type.StartsWith ("System.") && type.IndexOf ('.') == type.LastIndexOf ('.'))
				return type.Substring (7);
			break;
		}

		if (isArray)
			retval = retval + "[]";
		
		return retval;
	}
}

//
// The Ecma documentation provider:
//
// It populates a tree with contents during the help assembly
//
public class EcmaProvider : Provider {
	string basedir;

	public EcmaProvider (string base_directory)
	{
		basedir = base_directory;

		if (!Directory.Exists (basedir))
			throw new FileNotFoundException (String.Format ("The directory `{0}' does not exist", basedir));
	}
	
	public override void PopulateTree (Tree tree)
	{
		string [] namespaces = Directory.GetDirectories (basedir);

		foreach (string ns in namespaces){
			string [] files = Directory.GetFiles (ns);
			Node ns_node = null;
			string tn = null;
			
			foreach (string file in files){
				if (!file.EndsWith (".xml"))
					continue;

				if (ns_node == null){
					tn = Path.GetFileName (ns);
					Console.Error.WriteLine ("Processing namespace {0}", tn);
					ns_node = tree.LookupNode (tn, "N:" + tn);
					string ns_summary_file = basedir + "/" + tn + ".xml";
					if (File.Exists (ns_summary_file))
						ns_node.tree.HelpSource.PackFile (ns_summary_file,tn+".xml");
				}
				Console.Error.WriteLine ("    Processing input file {0}", Path.GetFileName (file));

				PopulateClass (tn, ns_node, file);
			}
		}
	}

	struct TypeInfo {
		public string type_name;
		public string type_full;
		public string type_kind;
		public XmlNode type_doc;

		public TypeInfo (string k, string f, string s, XmlNode n)
		{
			type_name = s;
			type_doc = n;
			type_kind = k;
			type_full = f;
		}
	}
	
	//
	// Packs a file with the summary data
	//
	public override void CloseTree (HelpSource hs, Tree tree)
	{
		foreach (DictionaryEntry de in class_summaries){
			XmlDocument doc = new XmlDocument ();
			string ns = (string) de.Key;
			ArrayList list = (ArrayList) de.Value;

			XmlElement elements = doc.CreateElement ("elements");
			doc.AppendChild (elements);
			
			string file = "xml.summary." + ns;
			Console.Error.WriteLine ("Have {0} elements in the {1}", list.Count, ns);
			foreach (TypeInfo p in list){
				XmlElement e = null;
				
				switch (p.type_kind){
				case "Class":
					e = doc.CreateElement ("class"); 
					break;
					
				case "Enumeration":
					e = doc.CreateElement ("enum");
					break;
					
				case "Structure":
					e = doc.CreateElement ("struct");
					break;
					
				case "Delegate":
					e = doc.CreateElement ("delegate");
					break;
					
				case "Interface":
					e = doc.CreateElement ("interface");
					break;
				}
				
				e.SetAttribute ("name", p.type_name);
				e.SetAttribute ("fullname", p.type_full);
				XmlNode copy = doc.ImportNode (p.type_doc, true);
				e.AppendChild (copy);
				elements.AppendChild (e);
			}
			hs.PackXml ("xml.summary." + ns, doc);
		}
	}
	       
	static Hashtable class_summaries = new Hashtable ();
					     
	XmlDocument doc;
	
	void PopulateClass (string ns, Node ns_node, string file)
	{
		doc = new XmlDocument ();
		doc.Load (file);
		
		string name = EcmaDoc.GetClassName (doc);
		string kind = EcmaDoc.GetTypeKind (doc);
		string full = EcmaDoc.GetFullClassName (doc);

		Node class_node;
		string file_code = ns_node.tree.HelpSource.PackFile (file);

		XmlNode class_summary = doc.SelectSingleNode ("/Type/Docs/summary");
		ArrayList l = (ArrayList) class_summaries [ns];
		if (l == null)
			l = class_summaries [ns] = new ArrayList ();
		l.Add (new TypeInfo (kind, full, name, class_summary));
	       
		class_node = ns_node.LookupNode (String.Format ("{0} {1}", name, kind), "ecma:" + file_code + "#" + name + "/");

		if (kind == "Enumeration")
			return;

		if (kind == "Delegate")
			return;
		
		//
		// Always add the Members node
		//
		class_node.CreateNode ("Members", "*");

		PopulateMember (class_node, "Constructor", "Constructors");
		PopulateMember (class_node, "Method", "Methods");
		PopulateMember (class_node, "Property", "Properties");
		PopulateMember (class_node, "Field", "Fields");
		PopulateMember (class_node, "Event", "Events");
	}

	class NodeIndex {
		public XmlNode node;
		public int     index;

		public NodeIndex (XmlNode node, int index)
		{
			this.node = node;
			this.index = index;
		}
	}

	struct NodeCompare : IComparer {
		public int Compare (object a, object b)
		{
			NodeIndex na = (NodeIndex) a;
			NodeIndex nb = (NodeIndex) b;

			return String.Compare (na.node.Attributes ["MemberName"].InnerText,
					       nb.node.Attributes ["MemberName"].InnerText);
		}
	}

	static NodeCompare NodeComparer = new NodeCompare ();

	string GetMemberName (XmlNode node)
	{
		return node.Attributes ["MemberName"].InnerText;
	}
	
	//
	// Performs an XPath query on the document to extract the nodes for the various members
	// we also use some extra text to pluralize the caption
	//
	void PopulateMember (Node node, string text, string plural_text)
	{
		XmlNodeList list = doc.SelectNodes (String.Format ("/Type/Members/Member[MemberType=\"{0}\"]", text));
		int count = list.Count;
		
		if (count == 0)
			return;

		Node nodes_node;
		string key = text.Substring (0, 1);
		if (count == 1)
			nodes_node = node.CreateNode (text, key);
		else
			nodes_node = node.CreateNode (plural_text, key);

		NodeIndex [] node_array = new NodeIndex [count];
		int i = 0;
		foreach (XmlNode n in list)
			node_array [i] = new NodeIndex (n, i++);
			
		Array.Sort (node_array, NodeComparer);

		foreach (NodeIndex ni in node_array){
			string signature;

			switch (text){
			case "Field":
				signature = GetMemberName (ni.node);
				break;
				
			default:
				XmlNode signode = ni.node.SelectSingleNode ("MemberSignature[@Language='C#']");
			
				if (signode == null)
					signature = GetMemberName (ni.node);
				else
					signature = signode.Attributes ["Value"].InnerText;
				break;
			}
			
			signature = CleanSignature (ni.node, signature);

			nodes_node.CreateNode (signature, ni.index.ToString ());
		}
	}

	//
	// This routine should be improved
	//
	string CleanSignature (XmlNode member_node, string input)
	{
		string [] remove_prefixes = { "public ", "private ", "protected", "internal ", "static ", "virtual", "override" };
		int matches;
		
		do {
			matches = 0;
			foreach (string s in remove_prefixes){
				if (input.StartsWith (s)){
					string old = input;
					try {
						input = input.Substring (s.Length);
					} catch {
						Console.Error.WriteLine ("Exception caught: old: {0} len: {1}", old, s.Length);
					}
					matches++;
				}
			}
		} while (matches > 0);

		//
		// Now, remove the rest until we find the membername (
		//
		string member_name = member_node.Attributes ["MemberName"].InnerText;

		int idx = input.IndexOf (member_name);
		if (idx <= 0)
			return input;

		return input.Substring (idx);
	}
}

//
// A help source renders information at runtime
//
public class EcmaHelpSource : HelpSource {

	public EcmaHelpSource (string base_file, bool create) : base (base_file, create)
	{
	}

	public override string GetText (string url, out Node match_node)
	{
		match_node = null;
		if (url.StartsWith ("ecma:"))
			return GetTextFromUrl (url);

		return null;
	}

	string RenderFieldLookup (string member, ref Node type_node)
	{
		foreach (Node x in type_node.Nodes){
			if (x.Caption != "Fields")
				continue;
			
			foreach (Node field in x.Nodes){
				if (field.Caption != member)
					continue;

				type_node = field;

				return GetTextFromUrl (field.URL);
			}
		}
		return "";
	}
	
	//
	// This routine has to perform a lookup on a type.
	//
	// Example: T:System.Text.StringBuilder
	//
	// The prefix is the kind of opereation being requested (T:, E:, M: etc)
	// ns is the namespace being looked up
	// type is the type being requested
	//
	// This has to walk our toplevel (which is always a namespace)
	// And then the type space, and then walk further down depending on the request
	//
	public override string RenderTypeLookup (string prefix, string ns, string type, string member, out Node match_node)
	{
		string nsp = prefix + ns;
		
		foreach (Node ns_node in Tree.Nodes){
			string ns_node_namespace = ns_node.Element.Substring (2);
			
			if (ns_node_namespace != ns)
				continue;
			
			foreach (Node type_node in ns_node.Nodes){
				string element = type_node.Element;
				int pidx = element.IndexOf ("#");
				int sidx = element.IndexOf ("/");
				
				string cname = element.Substring (pidx + 1, sidx-pidx-1);
				
				//Console.WriteLine ("t:{0} cn:{1} p:{2}", type, cname, prefix);
				if (type == cname && prefix == "T:"){
					match_node = type_node;
					return GetTextFromUrl (type_node.URL);
				} else if (type.StartsWith (cname)){
					int p = cname.Length;

					match_node = type_node;
					if (type == cname){
						if (prefix == "F:"){
							return RenderFieldLookup (member, ref match_node);
						}
						
						return string.Format ("<b>Found, now should process the {0} component ({1})</b>", prefix, member);
					} else if (type [p] == '/'){
						//
						// This handles summaries
						//
						match_node = null;
						Console.Error.WriteLine ("FIXME: Should set the right node here");
						return GetTextFromUrl (type_node.URL + type.Substring (p + 1));
					}
				}
			}
			match_node = null;
			return String.Format ("<b>FOUND!</b><p><p><b>Namespace: <b>{0}<br><b>Type:</b> {1}", ns, type);
		}

		match_node = null;
		return null;
	}

	//
	// Renders information from the summary node
	//
	void RenderSummary (StringBuilder sb, XmlDocument doc, string match, string label)
	{
		XmlNodeList list = doc.SelectNodes ("/elements" + match);

		if (list.Count == 0)
			return;
		
		H.Header (sb, label);
		sb.Append ("<table border=1 cellpadding=10><tr><td>Name</td><td>Description</td></tr>");
		foreach (XmlNode n in list){
			sb.Append (String.Format ("<tr><td><a href=\"T:{0}\"/>{1}</a></td>",
						  n.Attributes ["fullname"].InnerText,
						  n.Attributes ["name"].InnerText));
			XmlNode s = n.SelectSingleNode ("summary");

			sb.Append ("<td>");
			sb.Append (Htmlize (s));
			sb.Append ("</td>");
		}
		
		sb.Append ("</table>");
	}

	static void RenderRemarks (StringBuilder sb, XmlNode remarks)
	{
		if (remarks == null)
			return;
		
		H.Header (sb, "Remarks");
		H.BodyText (sb, Htmlize (remarks));
	}

	public override string RenderNamespaceLookup (string nsurl, out Node match_node)
	{
		foreach (Node ns_node in Tree.Nodes){
			if (ns_node.Element != nsurl)
				continue;

			match_node = ns_node;
			StringBuilder sb = new StringBuilder ();
			string ns_name = nsurl.Substring (2);
			Stream intro = GetHelpStream (ns_name + ".xml");
			
			Stream summary = GetHelpStream ("xml.summary." + ns_name);
			if (summary == null)
				return null;
			
			XmlDocument summary_doc = new XmlDocument ();
			summary_doc = new XmlDocument ();
			summary_doc.Load (summary);

			H.RenderCaption (sb, String.Format ("{0} Namespace", ns_name));
			//
			// Summary
			//
			
			if (intro != null){
				XmlDocument intro_doc = new XmlDocument ();
				intro_doc.Load (intro);

				H.BodyText (sb, Htmlize (intro_doc.SelectSingleNode ("Namespace/Docs/summary")));
				sb.Append ("<p>");

				RenderRemarks (sb, intro_doc.SelectSingleNode ("Namespace/Docs/remarks"));
			}

			StringBuilder body = new StringBuilder ();
			RenderSummary (body, summary_doc, "/class", "Classes");
			RenderSummary (body, summary_doc, "/interface", "Interfaces");
			RenderSummary (body, summary_doc, "/struct", "Structs");
			RenderSummary (body, summary_doc, "/delegate", "Delegates");
			RenderSummary (body, summary_doc, "/enum", "Enumerations");
		
			H.BodyText (sb, body.ToString ());
			return sb.ToString ();
		}
		match_node = null;
		return null;
	}

							   
	string GetTextFromUrl (string url)
	{
		XmlDocument doc;
		string file, rest;

		//
		// Remove ecma:
		//
		url = url.Substring (5);
		file = GetFile (url, out rest);

		// Console.WriteLine ("Got [{0}] and [{1}]", file, rest);
		doc = GetDocument (this, file);

		sb.Length = 0;
		if (rest == "")
			return RenderClassSummary (doc, url);

		string [] nodes = rest.Split (new char [] {'/'});
		switch (nodes [0]){
		case "C":
			if (nodes.Length == 1)
				return RenderConstructorSummary (doc, url);
			else
				return RenderConstructor (doc, Int32.Parse (nodes [1]));
			
		case "M":
			if (nodes.Length == 1)
				return RenderMethodSummary (doc, url);
			else
				return RenderMethod (doc, Int32.Parse (nodes [1]));

		case "P":
			if (nodes.Length == 1)
				return RenderPropertySummary (doc, url);
			else
				return RenderProperty (doc, Int32.Parse (nodes [1]));

		case "F":
			if (nodes.Length == 1)
				return RenderFieldSummary (doc, url);
			else
				return RenderField (doc, Int32.Parse (nodes [1]));

		case "E":
			if (nodes.Length == 1)
				return RenderEventSummary (doc, url);
			else
				return RenderEvent (doc, Int32.Parse (nodes [1]));

		case "*":
			return RenderMemberSummary (doc, url);
		}

		return "Unknown url: " + url;
	}

	static XslTransform ecma_transform;

	static string Htmlize (XmlNode ecma_xml)
	{
		if (ecma_transform == null){
			ecma_transform = new XslTransform ();
			Assembly assembly = System.Reflection.Assembly.GetCallingAssembly ();
			Stream stream = assembly.GetManifestResourceStream ("mono-ecma.xsl");

			XmlReader xml_reader = new XmlTextReader (stream);
			ecma_transform.Load (xml_reader);
		}
		
		if (ecma_xml != null){
			XmlDocument new_root = new XmlDocument ();

			new_root.AppendChild (new_root.ImportNode (ecma_xml, true));

			new_root.Save ("/tmp/blah.xml");
			StringWriter output = new StringWriter ();
			ecma_transform.Transform (new_root, null, output);
			
			return output.ToString ();
		} else
			return "";
	}

	//
	// This takes one of the ecma urls, which look like this:
	// ecma:NUMERIC_ID#OPAQUE/REST
	//
	// NUMERIC_ID is the numeric ID assigned by the compressor
	// OPAQUE is opaque for node rendering (it typically contains T:System.Byte for example)
	// REST is the rest of the argument used to decode information
	//
	static string GetFile (string url, out string rest)
	{
		int pound = url.IndexOf ("#");
		int slash = url.IndexOf ("/");
		
		string fname = url.Substring (0, pound);
		rest = url.Substring (slash+1);

		return fname;
	}

	// This should have a little cache or something.
	static XmlDocument GetDocument (HelpSource hs, string fname)
	{
		Stream s = hs.GetHelpStream (fname);
		if (s == null){
			Console.Error.WriteLine ("Could not fetch document {0}", fname);
			return null;
		}
		
		XmlDocument doc = new XmlDocument ();

		doc.Load (s);
		
		return doc;
	}

	static XmlNode FindMemberNode (XmlDocument doc, string type, int pos)
	{
		XmlNodeList list = doc.SelectNodes (String.Format ("/Type/Members/Member[MemberType=\"{0}\"]", type));

		return list [pos];
	}
	
#region Per-kind Rendering routines
	
	static StringBuilder sb = new StringBuilder ();
	
	static string RenderClassSummary (XmlDocument doc, string url)
	{
		string kind = EcmaDoc.GetTypeKind (doc);
		string cname = EcmaDoc.GetClassName (doc);
		
		H.RenderCaption (sb, String.Format ("{0} {1} <small>[C#]</small>", cname, kind));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (doc.SelectSingleNode ("Type/Docs/summary")));
		sb.Append ("<p>");

		//
		// Signature
		//
		XmlNode signode = doc.SelectSingleNode ("Type/TypeSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Thread safety
		//
		XmlNode ts = doc.SelectSingleNode ("Type/ThreadingSafetyStatement");
		if (ts != null){
			H.Header (sb, "Thread Safety");
			H.BodyText (sb, Htmlize (ts));
		}

		//
		// Remarks
		//
		XmlNode remarks = doc.SelectSingleNode ("Type/Docs/remarks");
		RenderRemarks (sb, remarks);

		//
		// Need to handle enums specially
		//
		if (kind == "Enumeration"){
			XmlNodeList members = doc.SelectNodes ("Type/Members/Member[MemberType=\"Field\"]");
			H.Header (sb, "Members");

			StringBuilder nested = new StringBuilder ();
			nested.Append ("<table border cellpadding=10 width=100%>");
			nested.Append ("<tr bgcolor=#bbbbbb><td><b>Member name</b></td><td><b>Description</b></td></tr>");
			foreach (XmlNode field in members){
				string name = field.Attributes ["MemberName"].InnerText;
				if (name == "value__")
					continue;
				nested.Append (
					String.Format ("<tr><td><b>{0}</b></td><td>{1}</td>",
						       name, Htmlize (field.SelectSingleNode ("Docs/summary"))));
			}
			
			nested.Append ("</table>");
			H.BodyText (sb, nested.ToString ());
		}
		
		H.Header (sb, "Requirements");
		string ns = EcmaDoc.GetClassNamespace (doc);
		H.BodyText (sb, String.Format ("<b>Namespace:</b> {0}<br>", ns));

		string ass = EcmaDoc.GetAssembly (doc);
		H.BodyText (sb, String.Format ("<b>Assembly:</b> {0} (in {1}.dll)<br>", ass, ass));

		H.Header (sb, "See Also");
		H.BodyText (sb,
			    String.Format ("<a href=\"T:{0}.{1}/*\">{1} Members</a> | <a href=\"N:{0}\">{0} Namespace</a>",
					   ns, cname));

		return sb.ToString ();
	}
	
	string RenderMemberSummary (XmlDocument doc, string kind, string kind_caption, string url, bool header, bool footer)
	{
		string cname = EcmaDoc.GetClassName (doc);

		if (url.EndsWith ("*")){
			url = url.Substring (0, url.Length-1) + kind [0];
		}
			
		if (header){
			H.RenderCaption (sb, String.Format ("{0} {1}", cname, kind_caption));
			H.BodyText (sb, String.Format (
					    "<p>The {0} of the {1} class are listed here.  For a complete list of {1} class members, see the " + 
					    "{1} Members topic.</p>", kind_caption.ToLower (), cname));
		}

		XmlNodeList methods = doc.SelectNodes (String.Format ("/Type/Members/Member[MemberType=\"{0}\"]", kind));

		//
		// First pull all the public methods
		//
		bool shown = false, found_protected = false;
		StringBuilder nested = new StringBuilder ();
		int pos = -1;
		foreach (XmlNode method in methods){
			pos++;
			
			XmlNode sig_node = method.SelectSingleNode ("MemberSignature[@Language='C#']");
			string sig = sig_node.Attributes ["Value"].InnerText;

			if (sig.StartsWith ("protected")){
				found_protected = true;
				continue;
			}

			if (!shown){
				H.Header (sb, "Public " + kind_caption);
				nested.Append ("<table border cellpadding=10 border=1 width=100%>");
				shown = true;
			}

			string name = method.Attributes ["MemberName"].InnerText;
			//
			// Horrible hack
			//
			if (name.StartsWith (".c"))
				name = String.Format ("{0} Constructor", cname);
			
			string desc = Htmlize (method.SelectSingleNode ("Docs/summary"));

			nested.Append (
				String.Format (
					"<tr><td><a href=\"source-id:{4}:ecma:{0}/{1}\">{2}</a></td><td>{3}</td></tr>",
					url, pos, name, desc, SourceID));
		}
		if (shown){
			nested.Append ("</table>");
			H.BodyText (sb, nested.ToString ());
			nested.Length = 0;
		}

		//
		// Repeat with protected
		//
		pos = -1;
		foreach (XmlNode method in methods){
			pos++;
			
			XmlNode sig_node = method.SelectSingleNode ("MemberSignature[@Language='C#']");
			string sig = sig_node.Attributes ["Value"].InnerText;

			if (!sig.StartsWith ("protected"))
				continue;

			if (!shown){
				H.Header (sb, "Protected methods");
				nested.Append ("<table border cellpadding=10 border=1 width=100%>");
				shown = true;
			}

			string name = method.Attributes ["MemberName"].InnerText;
			string desc = Htmlize (method.SelectSingleNode ("Docs/summary"));

			nested.Append (
				String.Format (
					"<tr><td><a href=\"source-id:{4}:ecma:{0}/{1}\">{2}</a></td><td>{3}</td></tr>",
					url, pos, name, desc, SourceID));
		}
		if (shown){
			nested.Append ("</table>");
			H.BodyText (sb, nested.ToString ());
		}

		if (footer){
			H.Header (sb, "See Also");

			string ns = EcmaDoc.GetClassNamespace (doc);
			H.BodyText (sb,
				    String.Format (
					    "<a href=\"T:{0}.{1}\">{1} {2}</a> | <a href=\"N:{0}\">{0} Namespace</a>",
					    ns, cname, EcmaDoc.GetTypeKind (doc)));
		}
		return sb.ToString ();
	}

	string RenderMethodSummary (XmlDocument doc, string url)
	{
		return RenderMemberSummary (doc, "Method", "Methods", url, true, true);
	}
	
	string RenderFieldSummary (XmlDocument doc, string url)
	{
		return RenderMemberSummary (doc, "Field", "Fields", url, true, true);
	}

	string RenderConstructorSummary (XmlDocument doc, string url)
	{
		return RenderMemberSummary (doc, "Constructor", "Constructors", url, true, true);
	}

	string RenderPropertySummary (XmlDocument doc, string url)
	{
		return RenderMemberSummary (doc, "Property", "Properties", url, true, true);
	}

	string RenderEventSummary (XmlDocument doc, string url)
	{
		return RenderMemberSummary (doc, "Event", "Events", url, true, true);
	}

	string RenderMemberSummary (XmlDocument doc, string url)
	{
		StringBuilder nested = new StringBuilder ();
		string ns = EcmaDoc.GetClassNamespace (doc);
		string cname = EcmaDoc.GetClassName (doc);

		H.RenderCaption (nested, String.Format ("{0} Members", cname));

		H.BodyText (nested, String.Format ("<a href=\"T:{0}.{1}\">{1} overview</a><p>", ns, cname));

		nested.Append (RenderMemberSummary (doc, "Constructor", "Constructors", url, false, false));
		sb.Length = 0;
		nested.Append (RenderMemberSummary (doc, "Field", "Fields", url, false, false));
		sb.Length = 0;
		nested.Append (RenderMemberSummary (doc, "Method", "Methods", url, false, false));
		sb.Length = 0;
		nested.Append (RenderMemberSummary (doc, "Property", "Properties", url, false, false));
		sb.Length = 0;
		nested.Append (RenderMemberSummary (doc, "Event", "Events", url, false, true));

		return nested.ToString ();
	}

	string RenderConstructor (XmlDocument doc, int pos)
	{
		XmlNode n = FindMemberNode (doc, "Constructor", pos);

		H.RenderCaption (sb, String.Format ("{0} Constructor {1} <small>[C#]</small>", EcmaDoc.GetClassName (doc), EcmaDoc.GetArguments (n)));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (n.SelectSingleNode ("Docs/summary")));
		sb.Append ("<p>");

		//
		// Signature
		//
		XmlNodeList parameters = n.SelectNodes ("Parameters/Parameter");
		
		//
		// TODO: later use the params information to "highlight" types in the signature
		//
		XmlNode signode = n.SelectSingleNode ("MemberSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Parameters
		//
		if (parameters.Count != 0){
			H.Header (sb, "Parameters");

			foreach (XmlNode par in parameters){
				string name = par.Attributes ["Name"].InnerText;
				XmlNode param_docs = n.SelectSingleNode (
					String.Format ("Docs/param[@name=\"{0}\"]", name));
				string pdoc = Htmlize (param_docs);
				H.BodyText (sb, String.Format ("<i>{0}</i><blockquote>{1}</blockquote>", name, pdoc));
			}
		}

		//
		// Exceptions
		//
		XmlNodeList exceptions = n.SelectNodes ("Docs/exception");
		if (exceptions != null) 
			RenderExceptionsList (sb, exceptions);

		//
		// Remarks
		//

		RenderRemarks (sb, n.SelectSingleNode ("Docs/remarks"));
		return sb.ToString ();
	}

	string RenderProperty (XmlDocument doc, int pos)
	{
		XmlNode n = FindMemberNode (doc, "Property", pos);

		H.RenderCaption (sb, String.Format ("{0} Property <small>[C#]</small>", n.Attributes ["MemberName"].InnerText));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (n.SelectSingleNode ("Docs/summary")));
		sb.Append ("<p>");

		//
		// TODO: later use the params information to "highlight" types in the signature
		//
		XmlNode signode = n.SelectSingleNode ("MemberSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Return value:
		//
		XmlNode return_value = n.SelectSingleNode ("Docs/returns");
		if (return_value != null){
			H.Header (sb, "Return Value");
			H.BodyText (sb, Htmlize (return_value));
		}

		//
		// Exceptions
		//
		XmlNodeList exceptions = n.SelectNodes ("Docs/exception");
		if (exceptions != null) 
			RenderExceptionsList (sb, exceptions);

		//
		// Remarks
		//
		RenderRemarks (sb, n.SelectSingleNode ("Docs/remarks"));
		
		return sb.ToString ();
	}

	void RenderException (StringBuilder sb, XmlNode exception)
	{
		sb.Append ("Exception: " + exception.InnerXml);
		
		sb.Append ("<tr>");

		sb.Append (String.Format ("<td><a href=\"T:{0}\">{0}</a></td>", exception.Attributes ["cref"])); 
		sb.Append (String.Format ("<td>{0}</td>", Htmlize (exception)));

		sb.Append ("<tr>");
	}

	void RenderExceptionsList (StringBuilder sb, XmlNodeList exceptions)
	{
		if (exceptions.Count == 0)
			return;

		StringBuilder nested = new StringBuilder ();

		H.Header (sb, "Exceptions");
				
		nested.Append ("<table border cellpadding=10 width=100%>");
		nested.Append ("<tr bgcolor=#bbbbbb><td><b>Exception Type</b></td><td><b>Condition</b></td></tr>");

		foreach (XmlNode ex in exceptions)
			RenderException (nested, ex);

		nested.Append ("</table>");

		H.BodyText (sb, nested.ToString ());
	}

	string RenderMethod (XmlDocument doc, int pos)
	{
		XmlNode n = FindMemberNode (doc, "Method", pos);

		H.RenderCaption (sb, String.Format ("{0} Method <small>[C#]</small>", n.Attributes ["MemberName"].InnerText));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (n.SelectSingleNode ("Docs/summary")));
		sb.Append ("<p>");

		//
		// Signature
		//
		XmlNodeList parameters = n.SelectNodes ("Parameters/Parameter");
		
		//
		// TODO: later use the params information to "highlight" types in the signature
		//
		XmlNode signode = n.SelectSingleNode ("MemberSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Parameters
		//
		if (parameters.Count != 0){
			H.Header (sb, "Parameters");

			foreach (XmlNode par in parameters){
				XmlAttribute parname = par.Attributes ["Name"];
				if (parname != null){
					string name = parname.InnerText;
					XmlNode param_docs = n.SelectSingleNode (
						String.Format ("Docs/param[@name=\"{0}\"]", name));
					string pdoc = Htmlize (param_docs);
					H.BodyText (sb, String.Format ("<i>{0}</i><blockquote>{1}</blockquote>", name, pdoc));
				} else {
					H.BodyText (sb, String.Format ("Can not find attribute Name in Method {0} Parameters/Parameter", pos));
				}
			}
		}

		//
		// Return value:
		//
		XmlNode return_value = n.SelectSingleNode ("Docs/returns");
		if (return_value != null){
			H.Header (sb, "Return Value");
			H.BodyText (sb, Htmlize (return_value));
		}

		//
		// Exceptions
		//
		XmlNodeList exceptions = n.SelectNodes ("Docs/exception");
		if (exceptions != null) 
			RenderExceptionsList (sb, exceptions);

		//
		// Remarks
		//
		RenderRemarks (sb, n.SelectSingleNode ("Docs/remarks"));
		
		return sb.ToString ();
	}

	string RenderField (XmlDocument doc, int pos)
	{
		XmlNode n = FindMemberNode (doc, "Field", pos);

		H.RenderCaption (sb, String.Format ("{0} Field <small>[C#]</small>", n.Attributes ["MemberName"].InnerText));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (n.SelectSingleNode ("Docs/summary")));
		sb.Append ("<p>");

		//
		// Signature
		//
		XmlNode signode = n.SelectSingleNode ("MemberSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Remarks
		//
		RenderRemarks (sb, n.SelectSingleNode ("Docs/remarks"));
		
		return sb.ToString ();
	}

	string RenderEvent (XmlDocument doc, int pos)
	{
		XmlNode n = FindMemberNode (doc, "Event", pos);

		H.RenderCaption (sb, String.Format ("{0} Event <small>[C#]</small>", n.Attributes ["MemberName"].InnerText));

		//
		// Summary
		//
		H.BodyText (sb, Htmlize (n.SelectSingleNode ("Docs/summary")));
		sb.Append ("<p>");

		//
		// Signature
		//
		XmlNode signode = n.SelectSingleNode ("MemberSignature[@Language='C#']");
		H.Signature (sb, signode.Attributes ["Value"].InnerText);

		//
		// Remarks
		//
		RenderRemarks (sb, n.SelectSingleNode ("Docs/remarks"));
		
		return sb.ToString ();
	}

#endregion
}
